@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2022 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,7 +35,6 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
 #ifndef SHUT_RD
@@ -76,7 +75,7 @@ NOEXPORT void connect_setup(CLI *);
 NOEXPORT int connect_init(CLI *, int);
 NOEXPORT int redirect(CLI *);
 NOEXPORT void print_bound_address(CLI *);
-NOEXPORT void reset(SOCKET, char *);
+NOEXPORT void reset(SOCKET, const char *);
 
 /* allocate local data structure for the new thread */
 CLI *alloc_client_session(SERVICE_OPTIONS *opt, SOCKET rfd, SOCKET wfd) {
@@ -202,7 +201,7 @@ void client_free(CLI *c) {
 #endif /* __GNUC__ */
 NOEXPORT void exec_connect_loop(CLI *c) {
     unsigned long long seq=0;
-    char *fresh_id=c->tls->id;
+    const char *fresh_id=c->tls->id;
     unsigned retry;
 
     do {
@@ -262,7 +261,8 @@ NOEXPORT void exec_connect_once(CLI *fresh_c) {
 #endif /* __GNUC__ */
 NOEXPORT void client_run(CLI *c) {
     jmp_buf exception_buffer, *exception_backup;
-    int err, rst;
+    int err, rst_remote, rst_local;
+    char *close_status;
 #ifndef USE_FORK
     int num;
 #endif
@@ -302,13 +302,27 @@ NOEXPORT void client_run(CLI *c) {
     }
     c->exception_pointer=exception_backup;
 
-    rst=err==1 && c->opt->option.reset;
+    /* plain socket is remote in the server mode */
+    rst_remote=err==1 && c->opt->option.reset &&
+        (!c->fatal_alert || !c->opt->option.client);
+    /* plain socket is local in the client mode */
+    rst_local=err==1 && c->opt->option.reset &&
+        (!c->fatal_alert || c->opt->option.client);
+    if(rst_remote && rst_local)
+        close_status="reset";
+    else if(rst_remote)
+        close_status="reset/closed";
+    else if(rst_local)
+        close_status="closed/reset";
+    else
+        close_status="closed";
     s_log(LOG_NOTICE,
         "Connection %s: %llu byte(s) sent to TLS, %llu byte(s) sent to socket",
-        rst ? "reset" : "closed",
+        close_status,
         (unsigned long long)c->ssl_bytes, (unsigned long long)c->sock_bytes);
 #ifdef USE_WIN32
-    if(capwin_hwnd && err==1 && InterlockedExchange(&capwin_connectivity, 0))
+    if(capwin_hwnd && (rst_remote || rst_local)
+            && InterlockedExchange(&capwin_connectivity, 0))
         PostMessage(capwin_hwnd, WM_CAPWIN_NET_DOWN, 0, 0);
 #endif
 
@@ -339,8 +353,11 @@ NOEXPORT void client_run(CLI *c) {
 
         /* cleanup the remote socket */
     if(c->remote_fd.fd!=INVALID_SOCKET) { /* remote socket initialized */
-        if(rst && c->remote_fd.is_socket) /* reset */
-            reset(c->remote_fd.fd, "linger (remote)");
+        if(rst_remote && c->remote_fd.is_socket) /* reset */
+            reset(c->remote_fd.fd, "remote_fd");
+#ifndef USE_UCONTEXT
+        set_nonblock(c->remote_fd.fd, 0);
+#endif
         closesocket(c->remote_fd.fd);
         s_log(LOG_DEBUG, "Remote descriptor (FD=%ld) closed",
             (long)c->remote_fd.fd);
@@ -350,16 +367,19 @@ NOEXPORT void client_run(CLI *c) {
         /* cleanup the local socket */
     if(c->local_rfd.fd!=INVALID_SOCKET) { /* local socket initialized */
         if(c->local_rfd.fd==c->local_wfd.fd) {
-            if(rst && c->local_rfd.is_socket)
-                reset(c->local_rfd.fd, "linger (local)");
+            if(rst_local && c->local_rfd.is_socket)
+                reset(c->local_rfd.fd, "local_rfd/local_wfd");
+#ifndef USE_UCONTEXT
+            set_nonblock(c->local_rfd.fd, 0);
+#endif
             closesocket(c->local_rfd.fd);
             s_log(LOG_DEBUG, "Local descriptor (FD=%ld) closed",
                 (long)c->local_rfd.fd);
         } else { /* stdin/stdout */
-            if(rst && c->local_rfd.is_socket)
-                reset(c->local_rfd.fd, "linger (local_rfd)");
-            if(rst && c->local_wfd.is_socket)
-                reset(c->local_wfd.fd, "linger (local_wfd)");
+            if(rst_local && c->local_rfd.is_socket)
+                reset(c->local_rfd.fd, "local_rfd");
+            if(rst_local && c->local_wfd.is_socket)
+                reset(c->local_wfd.fd, "local_wfd");
         }
         c->local_rfd.fd=c->local_wfd.fd=INVALID_SOCKET;
     }
@@ -674,19 +694,18 @@ NOEXPORT void session_cache_retrieve(CLI *c) {
 #if OPENSSL_VERSION_NUMBER >= 0x10101000L
 NOEXPORT void print_tmp_key(SSL *s) {
     EVP_PKEY *key;
+    long tmp_key_found;
 
 #ifdef SSL_CTRL_GET_PEER_TMP_KEY
-    if (!SSL_get_peer_tmp_key(s, &key)) {
-        sslerror("SSL_get_peer_tmp_key");
-        return;
-    }
+    tmp_key_found=SSL_get_peer_tmp_key(s, &key);
 #else
-    if (!SSL_get_server_tmp_key(s, &key)) {
-        sslerror("SSL_get_server_tmp_key");
+    tmp_key_found=SSL_get_server_tmp_key(s, &key);
+#endif
+    if(!tmp_key_found) {
+        s_log(LOG_INFO, "No peer temporary key received");
         return;
     }
-#endif
-    switch (EVP_PKEY_id(key)) {
+    switch(EVP_PKEY_id(key)) {
     case EVP_PKEY_RSA:
         s_log(LOG_INFO, "Peer temporary key: RSA, %d bits", EVP_PKEY_bits(key));
         break;
@@ -715,7 +734,11 @@ NOEXPORT void print_tmp_key(SSL *s) {
 #endif /* OpenSSL 1.1.1 or later */
 
 NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+    const SSL_CIPHER *cipher;
+#else
     SSL_CIPHER *cipher;
+#endif /* OpenSSL 1.1.1 or later */
 #ifndef OPENSSL_NO_COMP
     const COMP_METHOD *compression, *expansion;
 #endif
@@ -728,7 +751,7 @@ NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
         SSL_session_reused(c->ssl) && !c->flag.psk ?
             "previous session reused" : "new session negotiated");
 
-    cipher=(SSL_CIPHER *)SSL_get_current_cipher(c->ssl);
+    cipher=SSL_get_current_cipher(c->ssl);
     s_log(LOG_INFO, "%s ciphersuite: %s (%d-bit encryption)",
         SSL_get_version(c->ssl), SSL_CIPHER_get_name(cipher),
         SSL_CIPHER_get_bits(cipher, NULL));
@@ -1017,14 +1040,14 @@ NOEXPORT void transfer(CLI *c) {
             case SSL_ERROR_NONE:
                 if(num==0) { /* nothing was written: ignore */
                     s_log(LOG_DEBUG, "SSL_write returned 0");
-                    break; /* do not reset the watchdog */
+                } else {
+                    memmove(c->sock_buff, c->sock_buff+num,
+                        c->sock_ptr-(size_t)num);
+                    c->sock_ptr-=(size_t)num;
+                    memset(c->sock_buff+c->sock_ptr, 0, (size_t)num); /* PPD */
+                    c->ssl_bytes+=(size_t)num;
+                    watchdog=0; /* data transferred -> reset the watchdog */
                 }
-                memmove(c->sock_buff, c->sock_buff+num,
-                    c->sock_ptr-(size_t)num);
-                c->sock_ptr-=(size_t)num;
-                memset(c->sock_buff+c->sock_ptr, 0, (size_t)num); /* paranoia */
-                c->ssl_bytes+=(size_t)num;
-                watchdog=0; /* reset the watchdog */
                 break;
             case SSL_ERROR_WANT_WRITE: /* buffered data? */
                 s_log(LOG_DEBUG, "SSL_write returned WANT_WRITE: retrying");
@@ -1075,10 +1098,10 @@ NOEXPORT void transfer(CLI *c) {
             case SSL_ERROR_NONE:
                 if(num==0) { /* nothing was read: ignore */
                     s_log(LOG_DEBUG, "SSL_read returned 0");
-                    break; /* do not reset the watchdog */
+                } else {
+                    c->ssl_ptr+=(size_t)num;
+                    watchdog=0; /* data transferred -> reset the watchdog */
                 }
-                c->ssl_ptr+=(size_t)num;
-                watchdog=0; /* reset the watchdog */
                 break;
             case SSL_ERROR_WANT_WRITE:
                 s_log(LOG_DEBUG, "SSL_read returned WANT_WRITE: retrying");
@@ -1396,8 +1419,6 @@ NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
 
 #else /* standard Unix version */
 
-extern char **environ;
-
 NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
     int fd[2], pid;
     char **env;
@@ -1459,6 +1480,7 @@ NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
 }
 
 char **env_alloc(CLI *c) {
+    extern char **environ;
     char **env=NULL, **p;
     unsigned n=0; /* (n+2) keeps the list NULL-terminated */
     char *name, host[40], port[6];
@@ -1480,7 +1502,8 @@ char **env_alloc(CLI *c) {
             env[n++]=str_dup("LD_PRELOAD_32=" LIBDIR "/libstunnel.so");
             env=str_realloc(env, (n+2)*sizeof(char *));
             env[n++]=str_dup("LD_PRELOAD_64=" LIBDIR "/" MACH64 "/libstunnel.so");
-#elif __osf /* for Tru64 _RLD_LIST is used instead */
+#elif defined(__osf) || defined(__osf__)
+            /* for Tru64 _RLD_LIST is used instead */
             env=str_realloc(env, (n+2)*sizeof(char *));
             env[n++]=str_dup("_RLD_LIST=" LIBDIR "/libstunnel.so:DEFAULT");
 #else
@@ -1675,28 +1698,16 @@ NOEXPORT int connect_init(CLI *c, int domain) {
 
     if(c->bind_addr) {
         /* setup bind_addr based on c->bind_addr */
-#if 0
-        /* IPv6 addresses converted from IPv4 cause timeouts */
-#ifdef USE_IPv6
-        if(c->bind_addr->sa.sa_family==AF_INET && domain==AF_INET6) {
-            /* convert the binding address from IPv4 to IPv6 */
-            memset(&bind_addr, 0, sizeof bind_addr);
-            bind_addr.in6.sin6_family=AF_INET6;
-            bind_addr.in6.sin6_port=c->bind_addr->in.sin_port;
-            /* address format example: ::ffff:192.0.2.128 */
-            memset((char *)&bind_addr.in6.sin6_addr+10, 0xff, 2);
-            memcpy((char *)&bind_addr.in6.sin6_addr+12,
-                &c->bind_addr->in.sin_addr, 4);
-        } else /* just make a local copy */
-#endif
-#endif
-            memcpy(&bind_addr, c->bind_addr, (size_t)addr_len(c->bind_addr));
+        memcpy(&bind_addr, c->bind_addr, (size_t)addr_len(c->bind_addr));
         /* perform the initial sanity checks before creating a socket */
         if(bind_addr.sa.sa_family!=domain) {
             s_log(LOG_DEBUG, "Cannot assign an AF=%d address an AF=%d socket",
                 bind_addr.sa.sa_family, domain);
             return 1; /* failure */
         }
+    } else {
+        /* only needed to avoid a warning in MSVC */
+        memset(&bind_addr, 0, sizeof bind_addr);
     }
 
     /* create a new socket */
@@ -1803,13 +1814,18 @@ NOEXPORT void print_bound_address(CLI *c) {
     str_free(txt);
 }
 
-NOEXPORT void reset(SOCKET fd, char *txt) { /* set lingering on a socket */
+NOEXPORT void reset(SOCKET fd, const char *txt) { /* set lingering on a socket */
     struct linger l;
 
+    s_log(LOG_DEBUG, "%s reset (FD=%ld)", txt, (long)fd);
     l.l_onoff=1;
     l.l_linger=0;
-    if(setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *)&l, sizeof l))
-        log_error(LOG_DEBUG, get_last_socket_error(), txt);
+    if(setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *)&l, sizeof l)) {
+        int err=get_last_socket_error();
+        char *message=str_printf("setsockopt(SO_LINGER) on %s", txt);
+        log_error(LOG_INFO, err, message);
+        str_free(message);
+    }
 }
 
 void throw_exception(CLI *c, int v) {
