@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2019 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,9 +35,12 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
+#if defined(__GNUC__) && defined(USE_WIN32)
+#pragma GCC diagnostic ignored "-Wformat"
+#endif /* defined(__GNUC__) && defined(USE_WIN32) */
+
 SERVICE_OPTIONS *current_section=NULL;
 
 /* try an empty passphrase first */
@@ -88,7 +91,7 @@ NOEXPORT int load_key_engine(SERVICE_OPTIONS *);
 NOEXPORT int cache_passwd_get_cb(char *, int, int, void *);
 NOEXPORT int cache_passwd_set_cb(char *, int, int, void *);
 NOEXPORT void set_prompt(const char *);
-NOEXPORT int ui_retry();
+NOEXPORT int ui_retry(void);
 
 /* session tickets */
 #if OPENSSL_VERSION_NUMBER >= 0x10101000L
@@ -126,7 +129,7 @@ NOEXPORT void cache_transfer(SSL_CTX *, const u_char, const long,
 NOEXPORT void info_callback(const SSL *, int, int);
 
 NOEXPORT void sslerror_queue(void);
-NOEXPORT void sslerror_log(unsigned long, const char *, int, char *);
+NOEXPORT void sslerror_log(unsigned long, const char *, int, const char *);
 
 /**************************************** initialize section->ctx */
 
@@ -137,19 +140,27 @@ typedef long SSL_OPTIONS_TYPE;
 #endif
 
 int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
-    /* create TLS context */
+    /* create a new TLS context */
 #if OPENSSL_VERSION_NUMBER>=0x10100000L
-    if(section->option.client)
-        section->ctx=SSL_CTX_new(TLS_client_method());
-    else /* server mode */
-        section->ctx=SSL_CTX_new(TLS_server_method());
-    if(!SSL_CTX_set_min_proto_version(section->ctx,
+#if OPENSSL_VERSION_NUMBER>=0x30000000L
+    section->ctx=SSL_CTX_new_ex(NULL,
+        EVP_default_properties_is_fips_enabled(NULL) ?
+            "fips=yes" : "provider!=fips",
+        section->option.client ?
+            TLS_client_method() : TLS_server_method());
+#else /* OPENSSL_VERSION_NUMBER<0x30000000L */
+    section->ctx=SSL_CTX_new(section->option.client ?
+        TLS_client_method() : TLS_server_method());
+#endif /* OPENSSL_VERSION_NUMBER>=0x30000000L */
+    if(section->min_proto_version &&
+            !SSL_CTX_set_min_proto_version(section->ctx,
             section->min_proto_version)) {
         s_log(LOG_ERR, "Failed to set the minimum protocol version 0x%X",
             section->min_proto_version);
         return 1; /* FAILED */
     }
-    if(!SSL_CTX_set_max_proto_version(section->ctx,
+    if(section->max_proto_version &&
+            !SSL_CTX_set_max_proto_version(section->ctx,
             section->max_proto_version)) {
         s_log(LOG_ERR, "Failed to set the maximum protocol version 0x%X",
             section->max_proto_version);
@@ -165,13 +176,32 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
         sslerror("SSL_CTX_new");
         return 1; /* FAILED */
     }
-    /* for callbacks */
+
+    /* allow callbacks to access their SERVICE_OPTIONS structure */
     if(!SSL_CTX_set_ex_data(section->ctx, index_ssl_ctx_opt, section)) {
         sslerror("SSL_CTX_set_ex_data");
         return 1; /* FAILED */
     }
     current_section=section; /* setup current section for callbacks */
 
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+    /* set the security level */
+    if(section->security_level>=0) {
+        /* set the user-specified value */
+        SSL_CTX_set_security_level(section->ctx, section->security_level);
+        s_log(LOG_INFO, "User-specified security level set: %d",
+            section->security_level);
+    } else if(SSL_CTX_get_security_level(section->ctx)<DEFAULT_SECURITY_LEVEL) {
+        /* set our default, as it is more secure than the OpenSSL default */
+        SSL_CTX_set_security_level(section->ctx, DEFAULT_SECURITY_LEVEL);
+        s_log(LOG_INFO, "stunnel default security level set: %d",
+            DEFAULT_SECURITY_LEVEL);
+    } else { /* our default is not more secure than the OpenSSL default */
+        s_log(LOG_INFO, "OpenSSL security level is used: %d",
+            SSL_CTX_get_security_level(section->ctx));
+    }
+#endif /* OpenSSL 1.1.0 or later */
+
     /* ciphers */
     if(section->cipher_list) {
         s_log(LOG_DEBUG, "Ciphers: %s", section->cipher_list);
@@ -194,6 +224,13 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
 
     /* TLS options: configure the stunnel defaults first */
     SSL_CTX_set_options(section->ctx, SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3);
+    /* no session ticket gets sent to the client at all in TLSv1.2
+       and below, but a stateful ticket will be sent in TLSv1.3 */
+#ifdef SSL_OP_NO_TICKET
+    if(!section->option.client && !section->option.session_resume) {
+        SSL_CTX_set_options(section->ctx, SSL_OP_NO_TICKET);
+    }
+#endif
 #ifdef SSL_OP_NO_COMPRESSION
     /* we implemented a better way to disable compression if needed */
     SSL_CTX_clear_options(section->ctx, SSL_OP_NO_COMPRESSION);
@@ -209,11 +246,12 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
 
     /* TLS options: log the configured values */
 #if OPENSSL_VERSION_NUMBER>=0x009080dfL
-    s_log(LOG_DEBUG, "TLS options: 0x%08lX (+0x%08lX, -0x%08lX)",
+    s_log(LOG_DEBUG,
+        "TLS options: 0x%" PRIX64 " (+0x%" PRIX64 ", -0x%" PRIX64 ")",
         SSL_CTX_get_options(section->ctx),
         section->ssl_options_set, section->ssl_options_clear);
 #else /* OpenSSL older than 0.9.8m */
-    s_log(LOG_DEBUG, "TLS options: 0x%08lX (+0x%08lX)",
+    s_log(LOG_DEBUG, "TLS options: 0x%" PRIX64 " (+0x%" PRIX64 ")",
         SSL_CTX_get_options(section->ctx), section->ssl_options_set);
 #endif /* OpenSSL 0.9.8m or later */
 
@@ -221,7 +259,7 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
     if(conf_init(section))
         return 1; /* FAILED */
 
-    /* mode */
+    /* setup mode of operation for the TLS state machine */
 #ifdef SSL_MODE_RELEASE_BUFFERS
     SSL_CTX_set_mode(section->ctx,
         SSL_MODE_ENABLE_PARTIAL_WRITE |
@@ -249,14 +287,27 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
         unsigned servname_len=(unsigned)strlen(section->servname);
         if(servname_len>SSL_MAX_SSL_SESSION_ID_LENGTH)
             servname_len=SSL_MAX_SSL_SESSION_ID_LENGTH;
+#ifndef OPENSSL_NO_TLS1_3
+        /* suppress all tickets (stateful and stateless) in TLSv1.3 */
+        if(!section->option.session_resume && !SSL_CTX_set_num_tickets(section->ctx, 0)) {
+            sslerror("SSL_CTX_set_num_tickets");
+            return 1; /* FAILED */
+        }
+#endif /* TLS 1.3 */
         if(!SSL_CTX_set_session_id_context(section->ctx,
                 (unsigned char *)section->servname, servname_len)) {
             sslerror("SSL_CTX_set_session_id_context");
             return 1; /* FAILED */
         }
     }
-    SSL_CTX_set_session_cache_mode(section->ctx,
-        SSL_SESS_CACHE_BOTH | SSL_SESS_CACHE_NO_INTERNAL_STORE);
+    if(section->option.session_resume) {
+        SSL_CTX_set_session_cache_mode(section->ctx,
+            SSL_SESS_CACHE_BOTH | SSL_SESS_CACHE_NO_INTERNAL_STORE);
+    } else {
+        SSL_CTX_set_session_cache_mode(section->ctx, SSL_SESS_CACHE_OFF);
+    }
+    s_log(LOG_INFO, "Session resumption %s", section->option.session_resume
+        ? "enabled" : "disabled");
     SSL_CTX_sess_set_cache_size(section->ctx, section->session_size);
     SSL_CTX_set_timeout(section->ctx, section->session_timeout);
     SSL_CTX_sess_set_new_cb(section->ctx, sess_new_cb);
@@ -274,18 +325,18 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
     if(verify_init(section))
         return 1; /* FAILED */
 
-    /* initialize the DH/ECDH key agreement in the server mode */
-    if(!section->option.client) {
+    /* initialize the DH/ECDH key agreement */
 #ifndef OPENSSL_NO_TLSEXT
+    if(!section->option.client)
         SSL_CTX_set_tlsext_servername_callback(section->ctx, servername_cb);
 #endif /* OPENSSL_NO_TLSEXT */
 #ifndef OPENSSL_NO_DH
-        dh_init(section); /* ignore the result (errors are not critical) */
+    dh_init(section); /* ignore the result (errors are not critical) */
 #endif /* OPENSSL_NO_DH */
 #ifndef OPENSSL_NO_ECDH
-        ecdh_init(section); /* ignore the result (errors are not critical) */
+    if(ecdh_init(section))
+        return 1; /* FAILED */
 #endif /* OPENSSL_NO_ECDH */
-    }
 
     return 0; /* OK */
 }
@@ -378,7 +429,11 @@ NOEXPORT int dh_init(SERVICE_OPTIONS *section) {
 
     section->option.dh_temp_params=0; /* disable by default */
 
-    /* check if DH is actually enabled for this section */
+    /* check if DH is needed for this section */
+    if(section->option.client) {
+        s_log(LOG_INFO, "DH initialization skipped: client section");
+        return 0; /* OK */
+    }
     ciphers=SSL_CTX_get_ciphers(section->ctx);
     if(!ciphers)
         return 1; /* ERROR (unlikely) */
@@ -395,7 +450,7 @@ NOEXPORT int dh_init(SERVICE_OPTIONS *section) {
         }
     }
     if(i==n) { /* no DH ciphers found */
-        s_log(LOG_INFO, "DH initialization not needed");
+        s_log(LOG_INFO, "DH initialization skipped: no DH ciphersuites");
         return 0; /* OK */
     }
 
@@ -479,9 +534,7 @@ NOEXPORT int SSL_CTX_set1_groups_list(SSL_CTX *ctx, char *list) {
 NOEXPORT int ecdh_init(SERVICE_OPTIONS *section) {
     s_log(LOG_DEBUG, "ECDH initialization");
     if(!SSL_CTX_set1_groups_list(section->ctx, section->curves)) {
-#if OPENSSL_VERSION_NUMBER >= 0x10101000L
-        sslerror("SSL_CTX_set1_groups_list");
-#endif /* OpenSSL version >= 1.1.1 */
+        s_log(LOG_ERR, "Invalid groups list in 'curves'");
         return 1; /* FAILED */
     }
     s_log(LOG_DEBUG, "ECDH initialized with curves %s", section->curves);
@@ -674,7 +727,7 @@ NOEXPORT unsigned psk_server_callback(SSL *ssl, const char *identity,
 }
 
 NOEXPORT int psk_compar(const void *a, const void *b) {
-    PSK_KEYS *x=*(PSK_KEYS **)a, *y=*(PSK_KEYS **)b;
+    const PSK_KEYS *x=*(PSK_KEYS *const*)a, *y=*(PSK_KEYS *const*)b;
 
 #if 0
     s_log(LOG_DEBUG, "PSK cmp: %s %s", x->identity, y->identity);
@@ -706,7 +759,7 @@ void psk_sort(PSK_TABLE *table, PSK_KEYS *head) {
 PSK_KEYS *psk_find(const PSK_TABLE *table, const char *identity) {
     PSK_KEYS key, *ptr=&key, **ret;
 
-    key.identity=(char *)identity;
+    key.identity=identity;
     ret=bsearch(&ptr,
         table->val, table->num, sizeof(PSK_KEYS *), psk_compar);
     return ret ? *ret : NULL;
@@ -842,27 +895,39 @@ NOEXPORT int load_key_file(SERVICE_OPTIONS *section) {
 #ifndef OPENSSL_NO_ENGINE
 
 NOEXPORT int load_cert_engine(SERVICE_OPTIONS *section) {
-    struct {
-        const char *id;
-        X509 *cert;
-    } parms;
+    X509 *cert;
 
     s_log(LOG_INFO, "Loading certificate from engine ID: %s", section->cert);
-    parms.id=section->cert;
-    parms.cert=NULL;
-    ENGINE_ctrl_cmd(section->engine, "LOAD_CERT_CTRL", 0, &parms, NULL, 1);
-    if(!parms.cert) {
-        sslerror("ENGINE_ctrl_cmd");
+    cert=engine_get_cert(section->engine, section->cert);
+    if(!cert)
         return 1; /* FAILED */
-    }
-    if(!SSL_CTX_use_certificate(section->ctx, parms.cert)) {
+    if(!SSL_CTX_use_certificate(section->ctx, cert)) {
         sslerror("SSL_CTX_use_certificate");
+        X509_free(cert);
         return 1; /* FAILED */
     }
     s_log(LOG_INFO, "Certificate loaded from engine ID: %s", section->cert);
+    X509_free(cert);
     return 0; /* OK */
 }
 
+UI_METHOD *ui_stunnel() {
+    static UI_METHOD *ui_method=NULL;
+
+    if(ui_method) /* already initialized */
+        return ui_method;
+    ui_method=UI_create_method("stunnel UI");
+    if(!ui_method) {
+        sslerror("UI_create_method");
+        return NULL;
+    }
+    UI_method_set_opener(ui_method, ui_get_opener());
+    UI_method_set_writer(ui_method, ui_get_writer());
+    UI_method_set_reader(ui_method, ui_get_reader());
+    UI_method_set_closer(ui_method, ui_get_closer());
+    return ui_method;
+}
+
 NOEXPORT int load_key_engine(SERVICE_OPTIONS *section) {
     int i;
     EVP_PKEY *pkey;
@@ -874,7 +939,7 @@ NOEXPORT int load_key_engine(SERVICE_OPTIONS *section) {
 
     for(i=0; i<3; i++) {
         pkey=ENGINE_load_private_key(section->engine, section->key,
-            UI_stunnel(), NULL);
+            ui_stunnel(), NULL);
         if(!pkey) {
             if(i<2 && ui_retry()) { /* wrong PIN */
                 sslerror_queue(); /* dump the error queue */
@@ -933,46 +998,87 @@ NOEXPORT int ui_retry() {
     unsigned long err=ERR_peek_error();
 
     switch(ERR_GET_LIB(err)) {
-    case ERR_LIB_ASN1:
-        return 1;
-    case ERR_LIB_PKCS12:
-        switch(ERR_GET_REASON(err)) {
-        case PKCS12_R_MAC_VERIFY_FAILURE:
-            return 1;
-        default:
-            return 0;
-        }
-    case ERR_LIB_EVP:
+    case ERR_LIB_EVP: /* 6 */
         switch(ERR_GET_REASON(err)) {
         case EVP_R_BAD_DECRYPT:
             return 1;
         default:
+            s_log(LOG_ERR, "Unhandled ERR_LIB_EVP error reason: %d",
+                ERR_GET_REASON(err));
             return 0;
         }
-    case ERR_LIB_PEM:
+    case ERR_LIB_PEM: /* 9 */
         switch(ERR_GET_REASON(err)) {
         case PEM_R_BAD_PASSWORD_READ:
+        case PEM_R_BAD_DECRYPT:
             return 1;
         default:
+            s_log(LOG_ERR, "Unhandled ERR_LIB_PEM error reason: %d",
+                ERR_GET_REASON(err));
             return 0;
         }
-    case ERR_LIB_UI:
+    case ERR_LIB_ASN1: /* 13 */
+        return 1;
+    case ERR_LIB_PKCS12: /* 35 */
+        switch(ERR_GET_REASON(err)) {
+        case PKCS12_R_MAC_VERIFY_FAILURE:
+            return 1;
+        default:
+            s_log(LOG_ERR, "Unhandled ERR_LIB_PKCS12 error reason: %d",
+                ERR_GET_REASON(err));
+            return 0;
+        }
+#ifdef ERR_LIB_DSO /* 37 */
+    case ERR_LIB_DSO:
+        return 1;
+#endif
+    case ERR_LIB_UI: /* 40 */
         switch(ERR_GET_REASON(err)) {
         case UI_R_RESULT_TOO_LARGE:
         case UI_R_RESULT_TOO_SMALL:
+#ifdef UI_R_PROCESSING_ERROR
+        case UI_R_PROCESSING_ERROR:
+#endif
             return 1;
         default:
+            s_log(LOG_ERR, "Unhandled ERR_LIB_UI error reason: %d",
+                ERR_GET_REASON(err));
             return 0;
         }
-    case ERR_LIB_USER: /* PKCS#11 hacks */
+#ifdef ERR_LIB_OSSL_STORE
+    case ERR_LIB_OSSL_STORE: /* 44 - added in OpenSSL 1.1.1 */
+        switch(ERR_GET_REASON(err)) {
+        case OSSL_STORE_R_BAD_PASSWORD_READ:
+            return 1;
+        default:
+            s_log(LOG_ERR, "Unhandled ERR_LIB_OSSL_STORE error reason: %d",
+                ERR_GET_REASON(err));
+            return 0;
+        }
+#endif
+#ifdef ERR_LIB_PROV
+    case ERR_LIB_PROV: /* 57 - added in OpenSSL 3.0 */
+        switch(ERR_GET_REASON(err)) {
+        case PROV_R_BAD_DECRYPT:
+            return 1;
+        default:
+            s_log(LOG_ERR, "Unhandled ERR_LIB_PROV error reason: %d",
+                ERR_GET_REASON(err));
+            return 0;
+        }
+#endif
+    case ERR_LIB_USER: /* 128 - PKCS#11 hacks */
         switch(ERR_GET_REASON(err)) {
         case 7UL: /* CKR_ARGUMENTS_BAD */
         case 0xa0UL: /* CKR_PIN_INCORRECT */
             return 1;
         default:
+            s_log(LOG_ERR, "Unhandled ERR_LIB_USER error reason: %d",
+                ERR_GET_REASON(err));
             return 0;
         }
     default:
+        s_log(LOG_ERR, "Unhandled error library: %d", ERR_GET_LIB(err));
         return 0;
     }
 }
@@ -1115,12 +1221,12 @@ NOEXPORT int ssl_tlsext_ticket_key_cb(SSL *ssl, unsigned char *key_name,
         return -1;
     }
     if(c->opt->ticket_key->key_len == 16)
-        cipher = EVP_aes_128_cbc();
+        cipher=EVP_aes_128_cbc();
     else /* c->opt->ticket_key->key_len == 32 */
-        cipher = EVP_aes_256_cbc();
+        cipher=EVP_aes_256_cbc();
     if(enc) { /* create new session */
         /* EVP_CIPHER_iv_length() returns 16 for either cipher EVP_aes_128_cbc() or EVP_aes_256_cbc() */
-        iv_len = EVP_CIPHER_iv_length(cipher);
+        iv_len=EVP_CIPHER_iv_length(cipher);
         if(RAND_bytes(iv, iv_len) <= 0) { /* RAND_bytes error */
             s_log(LOG_ERR, "RAND_bytes failed");
             return -1;
@@ -1314,7 +1420,8 @@ NOEXPORT void cache_new(SSL *ssl, SSL_SESSION *sess) {
 
 NOEXPORT SSL_SESSION *cache_get(SSL *ssl,
         const unsigned char *key, int key_len) {
-    unsigned char *val=NULL, *val_tmp=NULL;
+    unsigned char *val=NULL;
+    const unsigned char *val_tmp=NULL;
     ssize_t val_len=0;
     SSL_SESSION *sess;
 
@@ -1323,11 +1430,7 @@ NOEXPORT SSL_SESSION *cache_get(SSL *ssl,
     if(!val)
         return NULL;
     val_tmp=val;
-    sess=d2i_SSL_SESSION(NULL,
-#if OPENSSL_VERSION_NUMBER>=0x0090707fL
-        (const unsigned char **)
-#endif /* OpenSSL version >= 0.9.7g */
-        &val_tmp, (long)val_len);
+    sess=d2i_SSL_SESSION(NULL, &val_tmp, (long)val_len);
     str_free(val);
     return sess;
 }
@@ -1434,8 +1537,9 @@ NOEXPORT void cache_transfer(SSL_CTX *ctx, const u_char type,
     len=recv(s, (void *)packet, sizeof(CACHE_PACKET), 0);
     closesocket(s);
     if(len<0) {
-        if(get_last_socket_error()==S_EWOULDBLOCK ||
-                get_last_socket_error()==S_EAGAIN)
+        int err=get_last_socket_error();
+
+        if(err==S_EWOULDBLOCK || (S_EWOULDBLOCK!=S_EAGAIN && err==S_EAGAIN))
             s_log(LOG_INFO, "cache_transfer: recv timeout");
         else
             sockerror("cache_transfer: recv");
@@ -1469,58 +1573,66 @@ NOEXPORT void info_callback(const SSL *ssl, int where, int ret) {
     CLI *c;
     SSL_CTX *ctx;
     const char *state_string;
-
-    c=SSL_get_ex_data((SSL *)ssl, index_ssl_cli);
-    if(c) {
 #if OPENSSL_VERSION_NUMBER>=0x10100000L
-        OSSL_HANDSHAKE_STATE state=SSL_get_state(ssl);
+    OSSL_HANDSHAKE_STATE state=SSL_get_state(ssl);
 #else
-        int state=SSL_get_state((SSL *)ssl);
+    int state=SSL_get_state((SSL *)ssl);
 #endif
 
+    c=SSL_get_ex_data(ssl, index_ssl_cli);
+    if(!c) {
+        s_log(LOG_ERR,
+            "INTERNAL ERROR: info_callback() called without CLI, state = %x",
+            state);
+        return;
+    }
 #if 0
-        s_log(LOG_DEBUG, "state = %x", state);
+    s_log(LOG_DEBUG, "state = %x", state);
 #endif
 
+        /* do not reset the TLS socket after a fatal alert */
+    if(where & SSL_CB_ALERT && !strcmp(SSL_alert_type_string(ret), "F"))
+        c->fatal_alert=1;
+
         /* log the client certificate request (if received) */
 #ifndef SSL3_ST_CR_CERT_REQ_A
-        if(state==TLS_ST_CR_CERT_REQ)
+    if(state==TLS_ST_CR_CERT_REQ)
 #else
-        if(state==SSL3_ST_CR_CERT_REQ_A)
+    if(state==SSL3_ST_CR_CERT_REQ_A)
 #endif
-            print_client_CA_list(SSL_get_client_CA_list((SSL *)ssl));
+        print_CA_list("Received trusted client CA",
+            SSL_get_client_CA_list(ssl));
 #ifndef SSL3_ST_CR_SRVR_DONE_A
-        if(state==TLS_ST_CR_SRVR_DONE)
+    if(state==TLS_ST_CR_SRVR_DONE)
 #else
-        if(state==SSL3_ST_CR_SRVR_DONE_A)
+    if(state==SSL3_ST_CR_SRVR_DONE_A)
 #endif
-            if(!SSL_get_client_CA_list((SSL *)ssl))
-                s_log(LOG_INFO, "Client certificate not requested");
+        if(!SSL_get_client_CA_list(ssl))
+            s_log(LOG_INFO, "Client certificate not requested");
 
-        /* prevent renegotiation DoS attack */
-        if((where&SSL_CB_HANDSHAKE_DONE)
-                && c->reneg_state==RENEG_INIT) {
-            /* first (initial) handshake was completed, remember this,
-             * so that further renegotiation attempts can be detected */
-            c->reneg_state=RENEG_ESTABLISHED;
-        } else if((where&SSL_CB_ACCEPT_LOOP)
-                && c->reneg_state==RENEG_ESTABLISHED) {
+    /* prevent renegotiation DoS attack */
+    if((where&SSL_CB_HANDSHAKE_DONE)
+            && c->reneg_state==RENEG_INIT) {
+        /* first (initial) handshake was completed, remember this,
+         * so that further renegotiation attempts can be detected */
+        c->reneg_state=RENEG_ESTABLISHED;
+    } else if((where&SSL_CB_ACCEPT_LOOP)
+            && c->reneg_state==RENEG_ESTABLISHED) {
 #ifndef SSL3_ST_SR_CLNT_HELLO_A
-            if(state==TLS_ST_SR_CLNT_HELLO) {
+        if(state==TLS_ST_SR_CLNT_HELLO) {
 #else
-            if(state==SSL3_ST_SR_CLNT_HELLO_A
-                    || state==SSL23_ST_SR_CLNT_HELLO_A) {
+        if(state==SSL3_ST_SR_CLNT_HELLO_A
+                || state==SSL23_ST_SR_CLNT_HELLO_A) {
 #endif
-                /* client hello received after initial handshake,
-                 * this means renegotiation -> mark it */
-                c->reneg_state=RENEG_DETECTED;
-            }
+            /* client hello received after initial handshake,
+             * this means renegotiation -> mark it */
+            c->reneg_state=RENEG_DETECTED;
         }
-
-        if(c->opt->log_level<LOG_DEBUG) /* performance optimization */
-            return;
     }
 
+    if(c->opt->log_level<LOG_DEBUG)
+        return; /* performance optimization: skip logging debug info */
+
     if(where & SSL_CB_LOOP) {
         state_string=SSL_state_string_long(ssl);
         if(strcmp(state_string, "unknown state"))
@@ -1534,7 +1646,7 @@ NOEXPORT void info_callback(const SSL *ssl, int where, int ret) {
             SSL_alert_type_string_long(ret),
             SSL_alert_desc_string_long(ret));
     } else if(where==SSL_CB_HANDSHAKE_DONE) {
-        ctx=SSL_get_SSL_CTX((SSL *)ssl);
+        ctx=SSL_get_SSL_CTX(ssl);
         if(c->opt->option.client) {
             s_log(LOG_DEBUG, "%6ld client connect(s) requested",
                 SSL_CTX_sess_connect(ctx));
@@ -1571,7 +1683,7 @@ NOEXPORT void info_callback(const SSL *ssl, int where, int ret) {
 
 /**************************************** TLS error reporting */
 
-void sslerror(char *txt) { /* OpenSSL error handler */
+void sslerror(const char *txt) { /* OpenSSL error handler */
     unsigned long err;
     const char *file;
     int line;
@@ -1598,7 +1710,7 @@ NOEXPORT void sslerror_queue(void) { /* recursive dump of the error queue */
 }
 
 NOEXPORT void sslerror_log(unsigned long err,
-        const char *file, int line, char *txt) {
+        const char *file, int line, const char *txt) {
     char *str;
 
     str=str_alloc(256);
