@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,7 +35,6 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
 #ifdef USE_WIN32
@@ -74,27 +73,56 @@ DISK_FILE *file_open(char *name, FILE_MODE mode) {
         SetFilePointer(fh, 0, NULL, FILE_END);
 
     /* setup df structure */
-    df=str_alloc(sizeof df);
+    df=str_alloc(sizeof(DISK_FILE));
     df->fh=fh;
     return df;
 }
 
 #else /* USE_WIN32 */
 
-DISK_FILE *file_fdopen(int fd) {
+DISK_FILE *file_fdopen(int fd, FILE_MODE file_mode) {
     DISK_FILE *df;
+    FILE *f;
+    char *mode;
 
+    switch(fd) {
+    case 0:
+        f=stdin;
+        break;
+    case 1:
+        f=stdout;
+        break;
+    case 2:
+        f=stderr;
+        break;
+    default:
+        switch(file_mode) {
+        case FILE_MODE_READ:
+            mode="r";
+            break;
+        case FILE_MODE_APPEND:
+            mode="a";
+            break;
+        case FILE_MODE_OVERWRITE:
+            mode="w";
+            break;
+        default: /* invalid file_mode */
+            return NULL;
+        }
+        f=fdopen(fd, mode);
+    }
+    if(!f)
+        return NULL;
     df=str_alloc(sizeof(DISK_FILE));
-    df->fd=fd;
+    df->f=f;
     return df;
 }
 
-DISK_FILE *file_open(char *name, FILE_MODE mode) {
-    DISK_FILE *df;
+DISK_FILE *file_open(char *name, FILE_MODE file_mode) {
     int fd, flags;
 
     /* open file */
-    switch(mode) {
+    switch(file_mode) {
     case FILE_MODE_READ:
         flags=O_RDONLY;
         break;
@@ -104,25 +132,17 @@ DISK_FILE *file_open(char *name, FILE_MODE mode) {
     case FILE_MODE_OVERWRITE:
         flags=O_CREAT|O_WRONLY|O_TRUNC;
         break;
-    default: /* invalid mode */
+    default: /* invalid file_mode */
         return NULL;
     }
-#ifdef O_NONBLOCK
-    flags|=O_NONBLOCK;
-#elif defined O_NDELAY
-    flags|=O_NDELAY;
-#endif
 #ifdef O_CLOEXEC
     flags|=O_CLOEXEC;
 #endif /* O_CLOEXEC */
+    /* don't fopen() directly to prevent O_CLOEXEC race condition */
     fd=open(name, flags, 0640);
     if(fd==INVALID_SOCKET)
         return NULL;
-
-    /* setup df structure */
-    df=str_alloc(sizeof df);
-    df->fd=fd;
-    return df;
+    return file_fdopen(fd, file_mode);
 }
 
 #endif /* USE_WIN32 */
@@ -133,20 +153,18 @@ void file_close(DISK_FILE *df) {
 #ifdef USE_WIN32
     CloseHandle(df->fh);
 #else /* USE_WIN32 */
-    if(df->fd>2) /* never close stdin/stdout/stder */
-        close(df->fd);
+    if(fileno(df->f)>2) /* never close stdin/stdout/stder */
+        fclose(df->f);
 #endif /* USE_WIN32 */
     str_free(df);
 }
 
 ssize_t file_getline(DISK_FILE *df, char *line, int len) {
-    /* this version is really slow, but performance is not important here */
-    /* (no buffering is implemented) */
     ssize_t i;
 #ifdef USE_WIN32
     DWORD num;
 #else /* USE_WIN32 */
-    ssize_t num;
+    int c;
 #endif /* USE_WIN32 */
 
     if(!df) /* not opened */
@@ -155,15 +173,20 @@ ssize_t file_getline(DISK_FILE *df, char *line, int len) {
     for(i=0; i<len-1; i++) {
 #ifdef USE_WIN32
         ReadFile(df->fh, line+i, 1, &num, NULL);
-#else /* USE_WIN32 */
-        num=read(df->fd, line+i, 1);
-#endif /* USE_WIN32 */
         if(num!=1) { /* EOF */
-            if(i) /* any previously retrieved data */
-                break;
-            else
+            if(!i) /* no previously retrieved data */
                 return -1;
+            break; /* MSDOS-style last file line */
         }
+#else /* USE_WIN32 */
+        c=getc(df->f);
+        if(c==EOF) {
+            if(!i) /* no previously retrieved data */
+                return -1;
+            break; /* MSDOS-style last file line */
+        }
+        line[i]=(char)c;
+#endif /* USE_WIN32 */
         if(line[i]=='\n') /* LF */
             break;
         if(line[i]=='\r') /* CR */
@@ -173,30 +196,40 @@ ssize_t file_getline(DISK_FILE *df, char *line, int len) {
     return i;
 }
 
-ssize_t file_putline(DISK_FILE *df, char *line) {
-    char *buff;
-    size_t len;
+ssize_t file_putline_nonewline(DISK_FILE *df, char *line) {
+    /* used for fatal_debug() -> no str.c functions are allowed */
 #ifdef USE_WIN32
     DWORD num;
+
+    if(df)
+        WriteFile(df->fh, line, (DWORD)strlen(line), &num, NULL);
 #else /* USE_WIN32 */
-    ssize_t num;
+    FILE *f;
+    int num;
+
+    f=df ? df->f : stderr; /* no file -> write to stderr */
+    num=fputs(line, f);
+    fflush(f);
 #endif /* USE_WIN32 */
+    return (ssize_t)num;
+}
+
+ssize_t file_putline_newline(DISK_FILE *df, char *line) {
+    char *buff;
+    size_t len;
+    ssize_t num;
 
     len=strlen(line);
-    buff=str_alloc(len+2); /* +2 for CR+LF */
+    buff=str_alloc(len+3); /* +3 for CR+LF+NUL */
     strcpy(buff, line);
 #ifdef USE_WIN32
     buff[len++]='\r'; /* CR */
 #endif /* USE_WIN32 */
     buff[len++]='\n'; /* LF */
-#ifdef USE_WIN32
-    WriteFile(df->fh, buff, (DWORD)len, &num, NULL);
-#else /* USE_WIN32 */
-    /* no file -> write to stderr */
-    num=write(df ? df->fd : 2, buff, len);
-#endif /* USE_WIN32 */
+    buff[len]='\0'; /* NUL */
+    num=file_putline_nonewline(df, buff);
     str_free(buff);
-    return (ssize_t)num;
+    return num;
 }
 
 int file_permissions(const char *file_name) {
