@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2019 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -44,6 +44,14 @@
 #define USE_OS_THREADS
 #endif
 
+#ifdef USE_OS_THREADS
+#ifndef USE_WIN32
+    /* FIXME: waiting for a single terminate_pipe socket
+     * from multiple threads causes a deadlock on WIN32 */
+#define USE_TERMINATE_PIPE
+#endif /* USE_WIN32 */
+#endif /* USE_OS_THREADS */
+
 /**************************************** forward declarations */
 
 typedef struct tls_data_struct TLS_DATA;
@@ -118,7 +126,10 @@ typedef struct sockaddr_list {                          /* list of addresses */
 
 #ifndef OPENSSL_NO_COMP
 typedef enum {
-    COMP_NONE, COMP_DEFLATE, COMP_ZLIB
+    COMP_NONE,                           /* empty compression algorithms set */
+    COMP_DEFLATE,            /* default OpenSSL's compression algorithms set */
+    COMP_ZLIB,          /* additional historic ZLIB compression algorithm id */
+    STUNNEL_COMPS                   /* number of compression algorithms sets */
 } COMP_TYPE;
 #endif /* !defined(OPENSSL_NO_COMP) */
 
@@ -175,7 +186,7 @@ typedef struct servername_list_struct SERVERNAME_LIST;/* forward declaration */
 
 #ifndef OPENSSL_NO_PSK
 typedef struct psk_keys_struct {
-    char *identity;
+    const char *identity;                  /* the OpenSSL API requires const */
     unsigned char *key_val;
     unsigned key_len;
     struct psk_keys_struct *next;
@@ -219,10 +230,13 @@ typedef struct service_options_struct {
     SOCK_OPT *sock_opts;
 
         /* service-specific data for verify.c */
-    char *ca_dir;                              /* directory for hashed certs */
-    char *ca_file;                       /* file containing bunches of certs */
-    char *crl_dir;                              /* directory for hashed CRLs */
-    char *crl_file;                       /* file containing bunches of CRLs */
+#ifndef OPENSSL_NO_ENGINE
+    NAME_LIST *ca_engine;  /* engine-specific CA certificate identifier list */
+#endif
+    char *ca_dir;                    /* directory containing hashed CA certs */
+    char *ca_file;                  /* file containing concatenated CA certs */
+    char *crl_dir;                       /* directory containing hashed CRLs */
+    char *crl_file;                     /* file containing concatenated CRLs */
 #ifndef OPENSSL_NO_OCSP
     char *ocsp_url;
     unsigned long ocsp_flags;
@@ -240,9 +254,12 @@ typedef struct service_options_struct {
     char *cert;                                             /* cert filename */
     char *key;                               /* pem (priv key/cert) filename */
     long session_size, session_timeout;
-    long unsigned ssl_options_set;
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+    int security_level;
+#endif /* OpenSSL 1.1.0 or later */
+    uint64_t ssl_options_set;
 #if OPENSSL_VERSION_NUMBER>=0x009080dfL
-    long unsigned ssl_options_clear;
+    uint64_t ssl_options_clear;
 #endif /* OpenSSL 0.9.8m or later */
 #if OPENSSL_VERSION_NUMBER>=0x10100000L
     int min_proto_version, max_proto_version;
@@ -291,7 +308,8 @@ typedef struct service_options_struct {
     char *username;
 
         /* service-specific data for protocol.c */
-    char * protocol;
+    char *protocol;
+    NAME_LIST *protocol_header;
     char *protocol_host;
     char *protocol_domain;
     char *protocol_username;
@@ -319,6 +337,7 @@ typedef struct service_options_struct {
 #endif
         unsigned local:1;               /* outgoing interface specified */
         unsigned retry:1;               /* loop remote+program */
+        unsigned session_resume:1;      /* enable session resumption */
         unsigned sessiond:1;
 #ifndef USE_WIN32
         unsigned pty:1;
@@ -365,7 +384,7 @@ typedef union {
 } OPT_UNION;
 
 struct sock_opt_struct {
-    char *opt_str;
+    const char *opt_str;
     int  opt_level;
     int  opt_name;
     VAL_TYPE opt_type;
@@ -397,7 +416,7 @@ typedef struct disk_file {
 #ifdef USE_WIN32
     HANDLE fh;
 #else
-    int fd;
+    FILE *f;
 #endif
     /* the interface is prepared to easily implement buffering if needed */
 } DISK_FILE;
@@ -439,6 +458,7 @@ typedef struct client_data_struct {
     FD remote_fd;                                  /* remote file descriptor */
     unsigned long pid;                           /* PID of the local process */
     SOCKET fd;                                  /* temporary file descriptor */
+    int fatal_alert;                               /* received a fatal alert */
     RENEG_STATE reneg_state;         /* used to track renegotiation attempts */
     unsigned long long seq;          /* sequential thread number for logging */
     unsigned rr;    /* per-client sequential number for round-robin failover */
@@ -470,8 +490,6 @@ int main_configure(char *, char *);
 void main_cleanup(void);
 int drop_privileges(int);
 void daemon_loop(void);
-void unbind_ports(void);
-int bind_ports(void);
 void signal_post(uint8_t);
 #if !defined(USE_WIN32) && !defined(USE_OS2)
 void pid_status_hang(const char *);
@@ -480,14 +498,13 @@ void stunnel_info(int);
 
 /**************************************** prototypes for options.c */
 
-extern char *configuration_file;
 extern unsigned number_of_sections;
 
 int options_cmdline(char *, char *);
 int options_parse(CONF_TYPE);
 void options_defaults(void);
 void options_apply(void);
-void options_free(void);
+void options_free(int);
 
 void service_up_ref(SERVICE_OPTIONS *);
 void service_free(SERVICE_OPTIONS *);
@@ -497,10 +514,10 @@ void service_free(SERVICE_OPTIONS *);
 #ifndef USE_FORK
 void get_limits(void); /* setup global max_clients and max_fds */
 #endif
-SOCKET s_socket(int, int, int, int, char *);
-int s_pipe(int[2], int, char *);
-int s_socketpair(int, int, int, SOCKET[2], int, char *);
-SOCKET s_accept(SOCKET, struct sockaddr *, socklen_t *, int, char *);
+SOCKET s_socket(int, int, int, int, const char *);
+int s_pipe(int[2], int, const char *);
+int s_socketpair(int, int, int, SOCKET[2], int, const char *);
+SOCKET s_accept(SOCKET, struct sockaddr *, socklen_t *, int, const char *);
 void set_nonblock(SOCKET, unsigned long);
 
 /**************************************** prototypes for log.c */
@@ -517,8 +534,9 @@ void s_log(int, const char *, ...)
 #else
     ;
 #endif
+void s_vlog(int, const char *, va_list);
 char *log_id(CLI *);
-void fatal_debug(char *, const char *, int) NORETURN;
+void fatal_debug(const char *, const char *, int) NORETURN;
 #define fatal(a) fatal_debug((a), __FILE__, __LINE__)
 void ioerror(const char *);
 void sockerror(const char *);
@@ -547,6 +565,9 @@ int cron_init(void);
 extern int index_ssl_cli, index_ssl_ctx_opt;
 extern int index_session_authenticated, index_session_connect_address;
 
+int fips_default(void);
+int fips_available(void);
+void crypto_init(void);
 int ssl_init(void);
 int ssl_configure(GLOBAL_OPTIONS *);
 
@@ -564,13 +585,19 @@ int context_init(SERVICE_OPTIONS *);
 void psk_sort(PSK_TABLE *, PSK_KEYS *);
 PSK_KEYS *psk_find(const PSK_TABLE *, const char *);
 #endif /* !defined(OPENSSL_NO_PSK) */
+#ifndef OPENSSL_NO_ENGINE
+UI_METHOD *ui_stunnel(void);
+#endif /* !defined(OPENSSL_NO_ENGINE) */
 void print_session_id(SSL_SESSION *);
-void sslerror(char *);
+void sslerror(const char *);
 
 /**************************************** prototypes for verify.c */
 
 int verify_init(SERVICE_OPTIONS *);
-void print_client_CA_list(const STACK_OF(X509_NAME) *);
+#ifndef OPENSSL_NO_ENGINE
+X509 *engine_get_cert(ENGINE *, const char *);
+#endif
+void print_CA_list(const char *, const STACK_OF(X509_NAME) *);
 char *X509_NAME2text(X509_NAME *);
 
 /**************************************** prototypes for network.c */
@@ -623,6 +650,7 @@ void throw_exception(CLI *, int) NORETURN;
 int get_socket_error(const SOCKET);
 int s_connect(CLI *, SOCKADDR_UNION *, socklen_t);
 void s_write(CLI *, SOCKET fd, const void *, size_t);
+size_t s_read_eof(CLI *, SOCKET fd, void *, size_t);
 void s_read(CLI *, SOCKET fd, void *, size_t);
 void fd_putline(CLI *, SOCKET, const char *);
 char *fd_getline(CLI *, SOCKET);
@@ -634,10 +662,17 @@ void fd_printf(CLI *, SOCKET, const char *, ...)
     ;
 #endif
 void s_ssl_write(CLI *, const void *, int);
+size_t s_ssl_read_eof(CLI *, void *, int);
 void s_ssl_read(CLI *, void *, int);
 char *ssl_getstring(CLI *c);
 char *ssl_getline(CLI *c);
 void ssl_putline(CLI *c, const char *);
+void ssl_printf(CLI *, const char *, ...)
+#ifdef __GNUC__
+    __attribute__((format(printf, 2, 3)));
+#else
+    ;
+#endif
 
 /**************************************** prototype for protocol.c */
 
@@ -648,17 +683,23 @@ typedef enum {
     PROTOCOL_LATE
 } PHASE;
 
-char *protocol(CLI *, SERVICE_OPTIONS *opt, const PHASE);
+#ifdef USE_WIN32
+extern HWND capwin_hwnd;
+extern LONG capwin_connectivity;
+#endif
+
+const char *protocol(CLI *, SERVICE_OPTIONS *opt, const PHASE);
 
 /**************************************** prototypes for resolver.c */
 
-void resolver_init();
+void resolver_init(void);
+int use_ipv6(void);
 
 unsigned name2addr(SOCKADDR_UNION *, char *, int);
-unsigned hostport2addr(SOCKADDR_UNION *, char *, char *, int);
+unsigned hostport2addr(SOCKADDR_UNION *, char *, const char *, int);
 
 unsigned name2addrlist(SOCKADDR_LIST *, char *);
-unsigned hostport2addrlist(SOCKADDR_LIST *, char *, char *);
+unsigned hostport2addrlist(SOCKADDR_LIST *, char *, const char *);
 
 void addrlist_clear(SOCKADDR_LIST *, int);
 unsigned addrlist_dup(SOCKADDR_LIST *, const SOCKADDR_LIST *);
@@ -678,21 +719,8 @@ const char *s_gai_strerror(int);
 #endif
 
 #ifdef USE_WIN32
-
 /* rename some locally shadowed declarations */
 #define getnameinfo     local_getnameinfo
-
-#ifndef _WIN32_WCE
-typedef int (CALLBACK * GETADDRINFO) (const char *,
-    const char *, const struct addrinfo *, struct addrinfo **);
-typedef void (CALLBACK * FREEADDRINFO) (struct addrinfo *);
-typedef int (CALLBACK * GETNAMEINFO) (const struct sockaddr *, socklen_t,
-    char *, size_t, char *, size_t, int);
-extern GETADDRINFO s_getaddrinfo;
-extern FREEADDRINFO s_freeaddrinfo;
-extern GETNAMEINFO s_getnameinfo;
-#endif /* ! _WIN32_WCE */
-
 #endif /* USE_WIN32 */
 
 int getnameinfo(const struct sockaddr *, socklen_t,
@@ -717,9 +745,11 @@ struct CRYPTO_dynlock_value {
 #ifdef USE_WIN32
     CRITICAL_SECTION critical_section;
 #endif
+#ifdef DEBUG_LOCKS
     const char *init_file, *read_lock_file, *write_lock_file,
         *unlock_file, *destroy_file;
     int init_line, read_lock_line, write_lock_line, unlock_line, destroy_line;
+#endif
 };
 
 typedef struct CRYPTO_dynlock_value CRYPTO_RWLOCK;
@@ -798,12 +828,13 @@ void ignore_value(void *);
 /**************************************** prototypes for file.c */
 
 #ifndef USE_WIN32
-DISK_FILE *file_fdopen(int);
+DISK_FILE *file_fdopen(int, FILE_MODE mode);
 #endif
 DISK_FILE *file_open(char *, FILE_MODE mode);
 void file_close(DISK_FILE *);
 ssize_t file_getline(DISK_FILE *, char *, int);
-ssize_t file_putline(DISK_FILE *, char *);
+ssize_t file_putline_nonewline(DISK_FILE *, char *);
+ssize_t file_putline_newline(DISK_FILE *, char *);
 int file_permissions(const char *);
 
 #ifdef USE_WIN32
@@ -813,18 +844,18 @@ LPSTR tstr2str(LPCTSTR);
 
 /**************************************** prototypes for libwrap.c */
 
-int libwrap_init();
+int libwrap_init(void);
 void libwrap_auth(CLI *);
 
 /**************************************** prototypes for tls.c */
 
 extern volatile int tls_initialized;
 
-void tls_init();
-TLS_DATA *tls_alloc(CLI *, TLS_DATA *, char *);
-void tls_cleanup();
+void tls_init(void);
+TLS_DATA *tls_alloc(CLI *, TLS_DATA *, const char *);
+void tls_cleanup(void);
 void tls_set(TLS_DATA *);
-TLS_DATA *tls_get();
+TLS_DATA *tls_get(void);
 
 /**************************************** prototypes for str.c */
 
@@ -836,11 +867,12 @@ struct tls_data_struct {
     size_t alloc_bytes, alloc_blocks;
     CLI *c;
     SERVICE_OPTIONS *opt;
-    char *id;
+    const char *id;
 };
 
-void str_init(TLS_DATA *);
-void str_cleanup(TLS_DATA *);
+void str_init(void);
+void str_thread_init(TLS_DATA *);
+void str_thread_cleanup(TLS_DATA *);
 char *str_dup_debug(const char *, const char *, int);
 #define str_dup(a) str_dup_debug((a), __FILE__, __LINE__)
 char *str_dup_detached_debug(const char *, const char *, int);
@@ -856,21 +888,29 @@ char *str_printf(const char *, ...)
 LPTSTR str_tprintf(LPCTSTR, ...);
 #endif
 
-void str_canary_init();
-void str_stats();
+void str_canary_init(void);
+void str_stats(void);
+
 void *str_alloc_debug(size_t, const char *, int);
 #define str_alloc(a) str_alloc_debug((a), __FILE__, __LINE__)
 void *str_alloc_detached_debug(size_t, const char *, int);
 #define str_alloc_detached(a) str_alloc_detached_debug((a), __FILE__, __LINE__)
+
 void *str_realloc_debug(void *, size_t, const char *, int);
 #define str_realloc(a, b) str_realloc_debug((a), (b), __FILE__, __LINE__)
 void *str_realloc_detached_debug(void *, size_t, const char *, int);
 #define str_realloc_detached(a, b) str_realloc_detached_debug((a), (b), __FILE__, __LINE__)
+
 void str_detach_debug(void *, const char *, int);
 #define str_detach(a) str_detach_debug((a), __FILE__, __LINE__)
+void str_detach_const_debug(const void *, const char *, int);
+#define str_detach_const(a) str_detach_const_debug((a), __FILE__, __LINE__)
+
 void str_free_debug(void *, const char *, int);
 #define str_free(a) str_free_debug((a), __FILE__, __LINE__), (a)=NULL
 #define str_free_expression(a) str_free_debug((a), __FILE__, __LINE__)
+void str_free_const_debug(const void *, const char *, int);
+#define str_free_const(a) str_free_const_debug((a), __FILE__, __LINE__), (a)=NULL
 
 void leak_table_utilization(void);
 
@@ -889,7 +929,10 @@ void message_box(LPCTSTR, const UINT);
 
 int ui_passwd_cb(char *, int, int, void *);
 #ifndef OPENSSL_NO_ENGINE
-UI_METHOD *UI_stunnel(void);
+int (*ui_get_opener(void)) (UI *);
+int (*ui_get_writer(void)) (UI *, UI_STRING *);
+int (*ui_get_reader(void)) (UI *, UI_STRING *);
+int (*ui_get_closer(void)) (UI *);
 #endif /* !defined(OPENSSL_NO_ENGINE) */
 
 #ifdef ICON_IMAGE
