@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,14 +35,35 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
+#ifdef _MSC_VER
+/*
+ * Disable MSVC C4996 warning:
+ * warning C4996: 'inet_addr': Use inet_pton() or InetPton() instead
+ * warning C4996: 'inet_ntoa': Use inet_ntop() or InetNtop() instead
+ * warning C4996: 'gethostbyname': Use getaddrinfo() or GetAddrInfoW() instead
+ */
+#pragma warning(disable: 4996)
+#endif
+
 /**************************************** prototypes */
 
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
+
+typedef int (CALLBACK * GETADDRINFO) (const char *,
+    const char *, const struct addrinfo *, struct addrinfo **);
+typedef void (CALLBACK * FREEADDRINFO) (struct addrinfo *);
+typedef int (CALLBACK * GETNAMEINFO) (const struct sockaddr *, socklen_t,
+    char *, size_t, char *, size_t, int);
+extern GETADDRINFO s_getaddrinfo;
+extern FREEADDRINFO s_freeaddrinfo;
+extern GETNAMEINFO s_getnameinfo;
+
 NOEXPORT int get_ipv6(LPTSTR);
-#endif
+
+#endif /* defined(USE_WIN32) && !defined(_WIN32_WCE) */
+
 NOEXPORT void addrlist2addr(SOCKADDR_UNION *, SOCKADDR_LIST *);
 NOEXPORT void addrlist_reset(SOCKADDR_LIST *);
 
@@ -134,6 +155,18 @@ NOEXPORT int get_ipv6(LPTSTR file) {
 #endif /* __GNUC__ */
 #endif
 
+int use_ipv6() {
+#if defined(USE_WIN32) && !defined(_WIN32_WCE)
+    return s_getaddrinfo != NULL;
+#else /* defined(USE_WIN32) && !defined(_WIN32_WCE) */
+#if defined(USE_IPv6)
+    return 1;
+#else /* defined(USE_IPv6) */
+    return 0;
+#endif /* defined(USE_IPv6) */
+#endif /* defined(USE_WIN32) && !defined(_WIN32_WCE) */
+}
+
 /**************************************** stunnel resolver API */
 
 unsigned name2addr(SOCKADDR_UNION *addr, char *name, int passive) {
@@ -151,7 +184,7 @@ unsigned name2addr(SOCKADDR_UNION *addr, char *name, int passive) {
 }
 
 unsigned hostport2addr(SOCKADDR_UNION *addr,
-        char *host_name, char *port_name, int passive) {
+        char *host_name, const char *port_name, int passive) {
     SOCKADDR_LIST *addr_list;
     unsigned num;
 
@@ -225,7 +258,7 @@ unsigned name2addrlist(SOCKADDR_LIST *addr_list, char *name) {
 }
 
 unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
-        char *host_name, char *port_name) {
+        char *host_name, const char *port_name) {
     struct addrinfo hints, *res, *cur;
     int err, retry=0;
     unsigned num;
@@ -255,7 +288,7 @@ unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
         }
         if(err==EAI_AGAIN && ++retry<=3) {
             s_log(LOG_DEBUG, "getaddrinfo: EAI_AGAIN received: retrying");
-            sleep(1);
+            s_poll_sleep(1, 0);
             continue;
         }
 #ifdef AI_ADDRCONFIG
@@ -320,7 +353,7 @@ unsigned addrlist_dup(SOCKADDR_LIST *dst, const SOCKADDR_LIST *src) {
 }
 
 unsigned addrlist_resolve(SOCKADDR_LIST *addr_list) {
-    unsigned num=0, rnd;
+    unsigned num=0, rnd=0;
     NAME_LIST *host;
 
     addrlist_reset(addr_list);
@@ -350,24 +383,30 @@ char *s_ntop(SOCKADDR_UNION *addr, socklen_t addrlen) {
     if(err) {
         s_log(LOG_ERR, "getnameinfo: %s", s_gai_strerror(err));
         retval=str_dup("unresolvable address");
-    } else
-        retval=str_printf("%s:%s", host, port);
+    } else if(!*port && !*host)
+        retval=str_dup("unnamed socket");
+    else
+        retval=str_printf("%s%s%s", host, *host && *port ? ":" : "", port);
     str_free(host);
     str_free(port);
     return retval;
 }
 
 socklen_t addr_len(const SOCKADDR_UNION *addr) {
-    if(addr->sa.sa_family==AF_INET)
+    switch(addr->sa.sa_family) {
+    case AF_UNSPEC: /* 0 */
+        return 0;
+    case AF_INET: /* 2 (almost universally) */
         return sizeof(struct sockaddr_in);
 #ifdef USE_IPv6
-    if(addr->sa.sa_family==AF_INET6)
+    case AF_INET6:
         return sizeof(struct sockaddr_in6);
 #endif
 #ifdef HAVE_STRUCT_SOCKADDR_UN
-    if(addr->sa.sa_family==AF_UNIX)
+    case AF_UNIX:
         return sizeof(struct sockaddr_un);
 #endif
+    }
     s_log(LOG_ERR, "INTERNAL ERROR: Unknown sa_family: %d",
         addr->sa.sa_family);
     return sizeof(SOCKADDR_UNION);
