@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -40,9 +40,17 @@
 #include <os2.h>
 #endif
 
-#include "common.h"
 #include "prototypes.h"
 
+/**************************************** prototypes */
+
+NOEXPORT void thread_id_init(void);
+NOEXPORT void locking_init(void);
+#ifndef USE_FORK
+CLI *thread_head=NULL;
+NOEXPORT void thread_list_add(CLI *);
+#endif
+
 /**************************************** thread ID callbacks */
 
 #ifdef USE_UCONTEXT
@@ -103,11 +111,11 @@ NOEXPORT void threadid_func(CRYPTO_THREADID *tid) {
 }
 #endif
 
-void thread_id_init(void) {
+NOEXPORT void thread_id_init() {
 #if OPENSSL_VERSION_NUMBER>=0x10000000L && OPENSSL_VERSION_NUMBER<0x10100000L
     CRYPTO_THREADID_set_callback(threadid_func);
 #endif
-#if OPENSSL_VERSION_NUMBER<0x10000000L || !defined(OPENSSL_NO_DEPRECATED)
+#if OPENSSL_VERSION_NUMBER<0x10000000L
     CRYPTO_set_id_callback(stunnel_thread_id);
 #endif
 }
@@ -122,36 +130,61 @@ void thread_id_init(void) {
 NOEXPORT void s_lock_init_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     pthread_rwlock_init(&lock->rwlock, NULL);
+#ifdef DEBUG_LOCKS
     lock->init_file=file;
     lock->init_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
 }
 
 NOEXPORT void s_read_lock_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     pthread_rwlock_rdlock(&lock->rwlock);
+#ifdef DEBUG_LOCKS
     lock->read_lock_file=file;
     lock->read_lock_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
 }
 
 NOEXPORT void s_write_lock_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     pthread_rwlock_wrlock(&lock->rwlock);
+#ifdef DEBUG_LOCKS
     lock->write_lock_file=file;
     lock->write_lock_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
 }
 
 NOEXPORT void s_unlock_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     pthread_rwlock_unlock(&lock->rwlock);
+#ifdef DEBUG_LOCKS
     lock->unlock_file=file;
     lock->unlock_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
 }
 
 NOEXPORT void s_lock_destroy_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     pthread_rwlock_destroy(&lock->rwlock);
+#ifdef DEBUG_LOCKS
     lock->destroy_file=file;
     lock->destroy_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
     str_free(lock);
 }
 
@@ -166,36 +199,61 @@ NOEXPORT void s_lock_destroy_debug(struct CRYPTO_dynlock_value *lock,
 NOEXPORT void s_lock_init_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     InitializeCriticalSection(&lock->critical_section);
+#ifdef DEBUG_LOCKS
     lock->init_file=file;
     lock->init_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
 }
 
 NOEXPORT void s_read_lock_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     EnterCriticalSection(&lock->critical_section);
+#ifdef DEBUG_LOCKS
     lock->read_lock_file=file;
     lock->read_lock_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
 }
 
 NOEXPORT void s_write_lock_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     EnterCriticalSection(&lock->critical_section);
+#ifdef DEBUG_LOCKS
     lock->write_lock_file=file;
     lock->write_lock_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
 }
 
 NOEXPORT void s_unlock_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     LeaveCriticalSection(&lock->critical_section);
+#ifdef DEBUG_LOCKS
     lock->unlock_file=file;
     lock->unlock_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
 }
 
 NOEXPORT void s_lock_destroy_debug(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     DeleteCriticalSection(&lock->critical_section);
+#ifdef DEBUG_LOCKS
     lock->destroy_file=file;
     lock->destroy_line=line;
+#else
+    (void)file; /* squash the unused parameter warning */
+    (void)line; /* squash the unused parameter warning */
+#endif
     str_free(lock);
 }
 
@@ -208,11 +266,12 @@ NOEXPORT int s_atomic_add(int *val, int amount, CRYPTO_RWLOCK *lock) {
 #if !defined(USE_OS_THREADS)
     /* no synchronization is needed */
     return *val+=amount;
-#elif defined(__GNUC__) && defined(__ATOMIC_ACQ_REL)
+#elif defined(__ATOMIC_ACQ_REL)
     if(__atomic_is_lock_free(sizeof *val, val))
         return __atomic_add_fetch(val, amount, __ATOMIC_ACQ_REL);
 #elif defined(_MSC_VER)
-    return InterlockedExchangeAdd(val, amount)+amount;
+    /* casting is safe, because sizeof(long)==sizeof(int) on Windows */
+    return InterlockedExchangeAdd((long *)val, amount)+amount;
 #endif
     CRYPTO_THREAD_write_lock(lock);
     ret=(*val+=amount);
@@ -256,7 +315,6 @@ NOEXPORT void s_dynlock_lock_cb(int mode, struct CRYPTO_dynlock_value *lock,
 NOEXPORT void s_dynlock_destroy_cb(struct CRYPTO_dynlock_value *lock,
         const char *file, int line) {
     s_lock_destroy_debug(lock, file, line);
-    str_free(lock);
 }
 
 NOEXPORT void s_locking_cb(int mode, int type, const char *file, int line) {
@@ -295,7 +353,6 @@ int CRYPTO_THREAD_unlock(CRYPTO_RWLOCK *lock) {
 
 void CRYPTO_THREAD_lock_free(CRYPTO_RWLOCK *lock) {
     s_lock_destroy_debug(lock, __FILE__, __LINE__);
-    str_free(lock);
 }
 
 #else /* USE_OS_THREADS */
@@ -332,7 +389,7 @@ int CRYPTO_atomic_add(int *val, int amount, int *ret, CRYPTO_RWLOCK *lock) {
 
 #endif /* OPENSSL_VERSION_NUMBER<0x10100004L */
 
-void locking_init(void) {
+NOEXPORT void locking_init() {
     size_t i;
 #if defined(USE_OS_THREADS) && OPENSSL_VERSION_NUMBER<0x10100004L
     size_t num;
@@ -458,6 +515,7 @@ int create_client(SOCKET ls, SOCKET s, CLI *arg) {
     context->context.uc_stack.ss_flags=0;
 
     makecontext(&context->context, (void(*)(void))client_thread, ARGC, arg);
+    thread_list_add(arg);
     s_log(LOG_DEBUG, "New context created");
     return 0;
 }
@@ -502,14 +560,25 @@ int create_client(SOCKET ls, SOCKET s, CLI *arg) {
 
 #ifdef USE_PTHREAD
 
+NOEXPORT void *dummy_thread(void *arg) {
+    pthread_exit(arg);
+    return arg;
+}
+
 int sthreads_init(void) {
+    pthread_t thread_id;
+
+    /* this is a workaround for NPTL threads failing to invoke
+     * pthread_exit() or pthread_cancel() from a chroot jail */
+    if(!pthread_create(&thread_id, NULL, dummy_thread, NULL))
+        pthread_join(thread_id, NULL);
+
     thread_id_init();
     locking_init();
     return 0;
 }
 
 int create_client(SOCKET ls, SOCKET s, CLI *arg) {
-    pthread_t thread;
     pthread_attr_t pth_attr;
     int error;
 #if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
@@ -528,22 +597,25 @@ int create_client(SOCKET ls, SOCKET s, CLI *arg) {
     pthread_sigmask(SIG_SETMASK, &new_set, &old_set); /* block signals */
 #endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
     pthread_attr_init(&pth_attr);
-    pthread_attr_setdetachstate(&pth_attr, PTHREAD_CREATE_DETACHED);
     pthread_attr_setstacksize(&pth_attr, arg->opt->stack_size);
-    error=pthread_create(&thread, &pth_attr, client_thread, arg);
+
+    CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_THREAD_LIST]);
+    error=pthread_create(&arg->thread_id, &pth_attr, client_thread, arg);
     pthread_attr_destroy(&pth_attr);
 #if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
     pthread_sigmask(SIG_SETMASK, &old_set, NULL); /* unblock signals */
 #endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
-
     if(error) {
         errno=error;
         ioerror("pthread_create");
+        CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
         str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
     }
+    thread_list_add(arg);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
     return 0;
 }
 
@@ -562,20 +634,22 @@ int sthreads_init(void) {
 }
 
 int create_client(SOCKET ls, SOCKET s, CLI *arg) {
-    HANDLE thread;
-
     (void)ls; /* this parameter is only used with USE_FORK */
     s_log(LOG_DEBUG, "Creating a new thread");
-    thread=(HANDLE)_beginthreadex(NULL, (unsigned)arg->opt->stack_size,
-        client_thread, arg, STACK_SIZE_PARAM_IS_A_RESERVATION, NULL);
-    if(!thread) {
+    CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_THREAD_LIST]);
+    arg->thread_id=(HANDLE)_beginthreadex(NULL,
+        (unsigned)arg->opt->stack_size, client_thread, arg,
+        STACK_SIZE_PARAM_IS_A_RESERVATION, NULL);
+    if(!arg->thread_id) {
         ioerror("_beginthreadex");
+        CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
         str_free(arg);
         if(s!=INVALID_SOCKET)
             closesocket(s);
         return -1;
     }
-    CloseHandle(thread);
+    thread_list_add(arg);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
     s_log(LOG_DEBUG, "New thread created");
     return 0;
 }
@@ -723,4 +797,14 @@ void stack_info(size_t stack_size, int init) { /* 1-initialize, 0-display */
 
 #endif /* DEBUG_STACK_SIZE */
 
+#ifndef USE_FORK
+NOEXPORT void thread_list_add(CLI *c) {
+    c->thread_next=thread_head;
+    c->thread_prev=NULL;
+    if(thread_head)
+        thread_head->thread_prev=c;
+    thread_head=c;
+}
+#endif /* !USE_FORK */
+
 /* end of sthreads.c */
