@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,12 +35,11 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
 volatile int tls_initialized=0;
 
-NOEXPORT void tls_platform_init();
+NOEXPORT void tls_platform_init(void);
 #if OPENSSL_VERSION_NUMBER<0x10100000L
 NOEXPORT void free_function(void *);
 #endif
@@ -59,22 +58,23 @@ void tls_init() {
     CRYPTO_set_mem_ex_functions(str_alloc_detached_debug,
         str_realloc_detached_debug, free_function);
 #endif
+    str_init();
 }
 
 /* this has to be the first function called by a new thread */
-TLS_DATA *tls_alloc(CLI *c, TLS_DATA *inherited, char *txt) {
+TLS_DATA *tls_alloc(CLI *c, TLS_DATA *inherited, const char *txt) {
     TLS_DATA *tls_data;
 
     if(inherited) { /* reuse the thread-local storage after fork() */
         tls_data=inherited;
-        str_free(tls_data->id);
+        str_free_const(tls_data->id);
     } else {
         tls_data=calloc(1, sizeof(TLS_DATA));
         if(!tls_data)
             fatal("Out of memory");
         if(c)
             c->tls=tls_data;
-        str_init(tls_data);
+        str_thread_init(tls_data);
         tls_data->c=c;
         tls_data->opt=c?c->opt:&service_options;
     }
@@ -84,10 +84,10 @@ TLS_DATA *tls_alloc(CLI *c, TLS_DATA *inherited, char *txt) {
     /* str.c functions can be used below this point */
     if(txt) {
         tls_data->id=str_dup(txt);
-        str_detach(tls_data->id); /* it is deallocated after str_stats() */
+        str_detach_const(tls_data->id); /* it is deallocated after str_stats() */
     } else if(c) {
         tls_data->id=log_id(c);
-        str_detach(tls_data->id); /* it is deallocated after str_stats() */
+        str_detach_const(tls_data->id); /* it is deallocated after str_stats() */
     }
 
     return tls_data;
@@ -100,8 +100,8 @@ void tls_cleanup() {
     tls_data=tls_get();
     if(!tls_data)
         return;
-    str_cleanup(tls_data);
-    str_free(tls_data->id); /* detached allocation */
+    str_thread_cleanup(tls_data);
+    str_free_const(tls_data->id); /* detached allocation */
     tls_set(NULL);
     free(tls_data);
 }
