@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,38 +35,14 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
 int main(int argc, char *argv[]) {
     static struct WSAData wsa_state;
-    TCHAR *c, stunnel_exe_path[MAX_PATH];
 
     tls_init(); /* initialize thread-local storage */
-
-    /* set current working directory and engine path */
-    GetModuleFileName(0, stunnel_exe_path, MAX_PATH);
-    c=_tcsrchr(stunnel_exe_path, TEXT('\\')); /* last backslash */
-    if(c) { /* found */
-        *c=TEXT('\0'); /* truncate the program name */
-        c=_tcsrchr(stunnel_exe_path, TEXT('\\')); /* previous backslash */
-        if(c && !_tcscmp(c+1, TEXT("bin")))
-            *c=TEXT('\0'); /* truncate "bin" */
-    }
-#ifndef _WIN32_WCE
-    if(!SetCurrentDirectory(stunnel_exe_path)) {
-        /* log to stderr, as s_log() is not initialized */
-        _ftprintf(stderr, TEXT("Cannot set directory to %s"),
-            stunnel_exe_path);
-        return 1;
-    }
-    /* try to enter the "config" subdirectory, ignore the result */
-    SetCurrentDirectory(TEXT("config"));
-#endif
-    _tputenv(str_tprintf(TEXT("OPENSSL_ENGINES=%s\\engines"),
-        stunnel_exe_path));
-
-    if(WSAStartup(MAKEWORD(1, 1), &wsa_state))
+    crypto_init(); /* initialize libcrypto */
+    if(WSAStartup(MAKEWORD(2, 2), &wsa_state))
         return 1;
     resolver_init();
     main_init();
@@ -130,9 +106,23 @@ int ui_passwd_cb(char *buf, int size, int rwflag, void *userdata) {
 }
 
 #ifndef OPENSSL_NO_ENGINE
-UI_METHOD *UI_stunnel() {
-    return UI_OpenSSL();
+
+int (*ui_get_opener()) (UI *) {
+    return UI_method_get_opener(UI_OpenSSL());
 }
+
+int (*ui_get_writer()) (UI *, UI_STRING *) {
+    return UI_method_get_writer(UI_OpenSSL());
+}
+
+int (*ui_get_reader()) (UI *, UI_STRING *) {
+    return UI_method_get_reader(UI_OpenSSL());
+}
+
+int (*ui_get_closer()) (UI *) {
+    return UI_method_get_closer(UI_OpenSSL());
+}
+
 #endif
 
 /* end of ui_win_cli.c */
