@@ -33,7 +33,6 @@ from plugin_collection import PluginCollection
 
 EXIT_SUCCESS = 0
 EXIT_FAILURE = 1
-EXIT_SKIP = 125
 
 RESULT_PATH = os.getcwd()
 DEFAULT_PROG = os.path.join(RESULT_PATH, "../src/stunnel")
@@ -62,6 +61,10 @@ RE_OPENSSL_VERSION = re.compile(
 RE_LINE_IDX = re.compile(r" ^ Hello \s+ (?P<idx> 0 | [1-9][0-9]* ) $ ", re.X)
 
 
+class UnsupportedOpenSSL(Exception):
+    """Unsupported version of OpenSSL"""
+
+
 @dataclasses.dataclass(frozen=True)
 class LogEvent():
     """The base class for an event."""
@@ -346,7 +349,7 @@ class TestLogs(PrintLogs):
 
 
     async def check_version(self, cmd_str: str, p_err: str) -> None:
-        """Check the version of stunnel and openssl"""
+        """Check the version of python, stunnel and openssl"""
         tag = "check_version"
         lines = p_err.splitlines()
         if not lines:
@@ -361,9 +364,13 @@ class TestLogs(PrintLogs):
             if match:
                 openssl_version = match.group("version")
         if not openssl_version:
-            raise Exception("Stunnel was compiled and run with various OpenSSL versions")
-        if openssl_version < "1.0.2":
-            raise Exception(f"OpenSSL version {openssl_version} is deprecated and not supported")
+            raise Exception("Stunnel was compiled and run with different OpenSSL versions")
+        """TLSv1.1 and TLSv1.2 available only with OpenSSL version 1.0.1 and later"""
+        if openssl_version < "1.0.1":
+            raise UnsupportedOpenSSL(f"OpenSSL version {openssl_version} is deprecated and not supported")
+        if not (sys.version_info.major == 3 and sys.version_info.minor >= 7):
+            raise Exception("Python 3.7 or higher is required.\n"
+                + "You are using Python {}.{}.".format(sys.version_info.major, sys.version_info.minor))
         if not stunnel_version:
             raise Exception(
                 f"Could not find the version line in the `{cmd_str}` output:\n"
@@ -492,10 +499,7 @@ class TestResult():
                 if result == "UNKNOWN":
                     result = "succeeded"
                 break
-        dots = "."
-        for dummy in range(70):
-            dots = dots + "."
-        self.logger.info(dots + result)
+        self.logger.info("." * 70 + " " + result)
         return result
 
 
@@ -621,9 +625,12 @@ class TestSuite(TestResult):
                         LogEvent(
                             etype="log",
                             level=20,
-                            log=f"[{tag}] Done with task '{service} output'"
+                            log=f"[{tag}] Waiting for an EOF on the '{service}_output' reader socket"
                         )
                     )
+                    line = await p_out.read(1)
+                    if line:
+                        raise Exception(f"Did not expect to read {line!r}")
                     return
 
                 line = data.decode("UTF-8").rstrip("\r\n")
@@ -646,6 +653,16 @@ class TestSuite(TestResult):
                             port=port
                         )
                     )
+                elif re.search("Accepting new connections", line):
+                    await self.cfg.mainq.put(
+                        StunnelEvent(
+                            etype="stunnel_event",
+                            level=10,
+                            log=f"[{tag}] Accepting new connections",
+                            service=service,
+                            port=0
+                        )
+                    )
                 elif re.search(r"Service \[inetd\] started", line):
                     await self.cfg.mainq.put(
                         StunnelEvent(
@@ -673,7 +690,7 @@ class TestSuite(TestResult):
                             level=30,
                             log=f"[{tag}] Stunnel '{service}' configuration failed",
                             service=service,
-                            port=0
+                            port=-1
                         )
                     )
 
@@ -709,13 +726,13 @@ class TestSuite(TestResult):
             )
         )
         self.cfg.children[Keys(pid=proc.pid, service=service)] = proc
-        asyncio.create_task(self.stunnel_output(proc.stderr, service))
+        self.cfg.tasks[f"{service}_output"] = asyncio.create_task(self.stunnel_output(proc.stderr, service))
 
 
     async def check_listening_port(self, port:int, service: str) -> int:
         """Raise exception if configuration failed."""
         tag = "check_listening_port"
-        if port == 0:
+        if port == -1:
             raise Exception(f"stunnel \'{service}\' failed")
         await self.cfg.mainq.put(
             LogEvent(
@@ -726,6 +743,13 @@ class TestSuite(TestResult):
         )
 
 
+    async def accepting_connections(self, port:int, service: str) -> int:
+        """If stunnel bound, expect log: Accepting new connections."""
+        await self.check_listening_port(port, service)
+        if int(port) > 0:
+            await self.expect_event(self.cfg.logsq, "stunnel_event")
+
+
     async def start_stunnel(self, cfgfile: pathlib.Path, service: str) -> int:
         """Launch the stunnel with the specified config file."""
         tag = "start_stunnel"
@@ -753,7 +777,7 @@ class TestSuite(TestResult):
             )
         )
         evt = await self.expect_event(self.cfg.logsq, "stunnel_event")
-        await self.check_listening_port(evt.port, evt.service)
+        await self.accepting_connections(evt.port, evt.service)
         return evt.port
 
 
@@ -778,7 +802,7 @@ class TestSuite(TestResult):
                     log=f"[{tag}] {line}"
                 )
             )
-        for key, dummy in self.cfg.children.items():
+        for key, _ in self.cfg.children.items():
             os.kill(key.pid, signal.SIGHUP)
             await self.cfg.mainq.put(
                 LogEvent(
@@ -958,17 +982,27 @@ class TestSuite(TestResult):
                             log=f"[{tag}] Closing down the server writer socket"
                         )
                     )
-                    server_writer.close()
-                    await server_writer.wait_closed()
-                    await self.cfg.mainq.put(
-                        ListenerClientEvent(
-                            etype="client_done",
-                            level=10,
-                            log=f"[{tag}] The 'listener' task closed a connection to the client",
-                            peer=peer,
-                            conns=self.conns
+                    try:
+                        server_writer.close()
+                        await server_writer.wait_closed()
+                    except ConnectionResetError as err:
+                        await self.cfg.mainq.put(
+                            LogEvent(
+                                etype="log",
+                                level=20,
+                                log=f"[{tag}] Handling {peer}: {err}"
+                            )
+                        )
+                    finally:
+                        await self.cfg.mainq.put(
+                            ListenerClientEvent(
+                                etype="client_done",
+                                level=10,
+                                log=f"[{tag}] The 'listener' task closed a connection to the client",
+                                peer=peer,
+                                conns=self.conns
+                            )
                         )
-                    )
 
 
     async def start_listener(self) -> int:
@@ -1097,7 +1131,7 @@ class TestSuite(TestResult):
             for key in children:
                 self.cfg.children.pop(key)
 
-            wait_res = await asyncio.gather(*waiters)
+            wait_res = await asyncio.gather(*waiters, return_exceptions=True)
             await self.cfg.mainq.put(
                 LogEvent(
                     etype="log",
@@ -1148,7 +1182,7 @@ class TestSuite(TestResult):
                                 log=f"[{tag}] - {err!r}"
                             )
                         )
-                    wait_res = await asyncio.gather(proc.wait())
+                    wait_res = await asyncio.gather(proc.wait(), return_exceptions=True)
                     await self.cfg.mainq.put(
                         LogEvent(
                             etype="log",
@@ -1374,7 +1408,7 @@ class StunnelAcceptConnect(TestSuite):
             await self.cfg.mainq.put(
                 LogEvent(
                     etype="fatal_event",
-                    level=20,
+                    level=30,
                     log=f"[{tag}] {err}",
                 )
             )
@@ -1399,6 +1433,10 @@ class ExpectedConfigurationFailure(StunnelAcceptConnect):
     async def check_listening_port(self, port:int, service: str) -> int:
         """Configuration failed as expected."""
 
+    async def accepting_connections(self, port:int, service: str) -> int:
+        """If stunnel bound, expect log: Accepting new connections."""
+        if int(port) > 0:
+            await self.expect_event(self.cfg.logsq, "stunnel_event")
 
 class ClientInetd(StunnelAcceptConnect):
     """Base class for inetd mode tests.
@@ -1416,6 +1454,12 @@ class ClientInetd(StunnelAcceptConnect):
         """You do not want stunnel to have any accept option."""
 
 
+    async def accepting_connections(self, port:int, service: str) -> int:
+        """If stunnel bound, expect log: Accepting new connections."""
+        if int(port) > 0:
+            await self.expect_event(self.cfg.logsq, "stunnel_event")
+
+
     async def run_stunnel(self, cfgfile: pathlib.Path, service: str) -> int:
         """Create the stunnel subprocess."""
         tag = "run_stunnel"
@@ -1438,7 +1482,7 @@ class ClientInetd(StunnelAcceptConnect):
             )
         )
         self.cfg.children[Keys(pid=proc.pid, service=service)] = proc
-        asyncio.create_task(self.stunnel_output(proc.stderr, service))
+        self.cfg.tasks[f"{service}_output"] = asyncio.create_task(self.stunnel_output(proc.stderr, service))
 
 
     async def get_io_stream(
@@ -1461,7 +1505,15 @@ class ClientConnectExec(TestSuite):
 
 
     async def check_listening_port(self, port:int, service: str) -> int:
-        """You do not want stunnel to have any accept option."""
+        """Raise exception if configuration failed."""
+        if port == -1:
+            raise Exception(f"stunnel \'{service}\' failed")
+
+
+    async def accepting_connections(self, port:int, service: str) -> int:
+        """Expect log: Accepting new connections."""
+        await self.check_listening_port(port, service)
+        await self.expect_event(self.cfg.logsq, "stunnel_event")
 
 
     async def socket_connected_cb(
@@ -1552,7 +1604,7 @@ class ClientConnectExec(TestSuite):
 
     async def start_connections(self, cfgfile: pathlib.Path, port: int) -> None:
         """Wait for all the connections to complete."""
-        for dummy in range(self.params.conn_num):
+        for _ in range(self.params.conn_num):
             await self.expect_event(self.cfg.logsq, "connection_done_event")
         await self.expect_event(self.cfg.logsq, "all_connections_event")
 
@@ -1599,6 +1651,14 @@ def parse_args() -> Config:
         help=f"the path to the test logs directory "
         f"(default: {DEFAULT_LOGS})",
     )
+    parser.add_argument(
+        "--libs",
+        type=pathlib.Path,
+        default="",
+        metavar="LIBRARY_PATH",
+        help="the path to the OpenSSL libraries "
+        "(default: NULL)",
+    )
     parser.add_argument(
         "--debug",
         type=int,
@@ -1609,7 +1669,11 @@ def parse_args() -> Config:
     )
     args = parser.parse_args()
     utf8_env = dict(os.environ)
-    utf8_env.update({"LC_ALL": "C.UTF-8", "LANGUAGE": ""})
+    # environment can only contain strings
+    utf8_env.update({
+        "LC_ALL": "C.UTF-8",
+        "LANGUAGE": "",
+        "LD_LIBRARY_PATH": str(args.libs)})
     if not os.path.isdir(args.logs):
         os.mkdir(args.logs)
     with os.scandir(args.logs) as entries:
@@ -1637,13 +1701,10 @@ def parse_args() -> Config:
 
 async def main() -> None:
     """Main program: parse arguments, prepare an environment, run tests."""
-    if not (sys.version_info.major == 3 and sys.version_info.minor >= 7):
-        print("Python 3.7 or higher is required.\n"
-            + "You are using Python {}.{}.".format(sys.version_info.major, sys.version_info.minor))
-        sys.exit(EXIT_SKIP)
     tag = "main"
-    omitted = False
     with parse_args() as cfg:
+
+        # Initialize the event processing infrastructure.
         try:
             rlogs = TestLogs(cfg)
             formats = "%(levelname)s: %(asctime)s: %(message)s"
@@ -1653,6 +1714,14 @@ async def main() -> None:
             slogs = TestLogs(cfg)
             formats = "%(message)s"
             slogger = slogs.setup_logger("summary", formats, cfg.summary, DEFAULT_LEVEL)
+        except Exception as err:  # pylint: disable=broad-except
+            # Logging is not available at this point.
+            print(err)
+            print("Framework initalization failed")
+            sys.exit(EXIT_FAILURE)
+
+        # Execute the tests.
+        try:
             await slogs.get_version(slogger)
             slogs.transcript_logs("summary", formats)
 
@@ -1664,6 +1733,15 @@ async def main() -> None:
                     log=f"[{tag}] Stunnel tests completed"
                 )
             )
+        except UnsupportedOpenSSL as err:
+            await cfg.mainq.put(
+                LogEvent(
+                    etype="finish_event",
+                    level=30,
+                    log=f"[{tag}] Unsupported OpenSSL: {err}"
+                )
+            )
+            print(err)
         except Exception as err:  # pylint: disable=broad-except
             await cfg.mainq.put(
                 LogEvent(
@@ -1673,19 +1751,24 @@ async def main() -> None:
                 )
             )
             print(err)
-            omitted = True
 
-        finally:
-            evt = await cfg.logsq.get()
-            log_error = 0 if evt.etype == "finish_event" else 1
+        # Report the results.
+        evt = await cfg.logsq.get()
+        if evt.etype == "finish_event":
             succeeded, failed, skipped = task.result()
-            slogger.info("Stunnel tests skipped" if omitted else
-                 f"\nSummary:\n   success: {succeeded}\n   fail: {failed}\n   skipped: {skipped}\n"
-                 + f"\nFile {cfg.results} " "done" if not log_error else
-                 "failed (Expected 'finish_event')")
-            if omitted:
-                sys.exit(EXIT_SKIP)
-            sys.exit(EXIT_SUCCESS if not failed else EXIT_FAILURE)
+            slogger.info("\nSummary:")
+            slogger.info("\tsucceeded: %s", succeeded)
+            slogger.info("\tfailed: %s", failed)
+            slogger.info("\tskipped: %s", skipped)
+            slogger.info("\nTesting results: %s", cfg.results)
+            if evt.level < 40: # not an error
+                if failed == 0:
+                    sys.exit(EXIT_SUCCESS)
+            else:
+                slogger.error("Failed (not all plugins were executed)")
+        else: # not synchronized -> stats may be misleading
+            slogger.error("Failed (expected 'finish_event')")
+        sys.exit(EXIT_FAILURE)
 
 
 if __name__ == "__main__":
